"""Stuff for handling errors."""

from colorama import Fore


def get_line_strings(source_code: str, line: int) -> str:
    """Get the line string from the source code."""
    lines = source_code.split("\n")

    return lines[line - 1]


def syntax_error(
    source_code: str, pos: tuple[int, int, int], error_message: str
) -> None:
    """Handle a syntax error."""
    line, col, length = pos  # Unpack the position tuple
    col -= length - 1  # Subtract the length of the error from the column
    print(pos)

    line_text = get_line_strings(source_code, line)  # Get a snippet of the code

    error_line = (
        f"{Fore.CYAN}{str(line).rjust(3)}"
        f" |{Fore.RESET} {line_text}"
    )  # Format the error line

    marker = (
        "^" * length
    )  # Under the error line, add a marker to show where the error is

    line_number_size = len(str(line).rjust(3))  # Get the size of the line number

    positioned_marker = marker.rjust(
        col + line_number_size + 2 + len(marker)
    )  # Position the marker

    formatted_marker = (
        f"{Fore.YELLOW}{positioned_marker}{Fore.RESET}"  # Format the marker
    )

    error_msg = f"{Fore.RED}SyntaxError{Fore.RESET}: {error_message}"  # Format the error message

    # pylint pls dont crash this time
    raise SyntaxError(f"{error_line}\n{formatted_marker}\n{error_msg}")



"""Lexer for the fronted."""

from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Tuple

from colorama import Fore

from eryx.utils.errors import syntax_error


class TokenType(Enum):
    """All token types in the language."""

    NUMBER = auto()
    IDENTIFIER = auto()
    STRING = auto()

    ASSIGNMENT_OPERATOR = auto()

    OPEN_PAREN = auto()
    CLOSE_PAREN = auto()
    OPEN_BRACE = auto()
    CLOSE_BRACE = auto()
    OPEN_BRACKET = auto()
    CLOSE_BRACKET = auto()

    DOUBLE_QUOTE = auto()

    BINARY_OPERATOR = auto()

    LET = auto()
    CONST = auto()
    FUNC = auto()
    IF = auto()
    ELSE = auto()
    RETURN = auto()
    ASSERT = auto()

    CLASS = auto()
    ENUM = auto()

    LOOP = auto()
    WHILE = auto()
    FOR = auto()
    IN = auto()
    BREAK = auto()
    CONTINUE = auto()

    IMPORT = auto()
    FROM = auto()
    AS = auto()

    EQUALS = auto()

    DEL = auto()

    COMMA = auto()
    COLON = auto()
    SEMICOLON = auto()
    DOT = auto()

    EOF = auto()


SINGLE_CHAR_TOKENS = {
    "(": TokenType.OPEN_PAREN,
    ")": TokenType.CLOSE_PAREN,
    "{": TokenType.OPEN_BRACE,
    "}": TokenType.CLOSE_BRACE,
    "[": TokenType.OPEN_BRACKET,
    "]": TokenType.CLOSE_BRACKET,
    "+": TokenType.BINARY_OPERATOR,
    "*": TokenType.BINARY_OPERATOR,
    "/": TokenType.BINARY_OPERATOR,
    "%": TokenType.BINARY_OPERATOR,
    "^": TokenType.BINARY_OPERATOR,
    ";": TokenType.SEMICOLON,
    ",": TokenType.COMMA,
    ":": TokenType.COLON,
    ".": TokenType.DOT,
    "=": TokenType.EQUALS,
    "<": TokenType.BINARY_OPERATOR,
    ">": TokenType.BINARY_OPERATOR,
    "&": TokenType.BINARY_OPERATOR,
    "|": TokenType.BINARY_OPERATOR,
}

DOUBLE_CHAR_TOKENS = {
    "==": TokenType.BINARY_OPERATOR,
    "!=": TokenType.BINARY_OPERATOR,
    "<=": TokenType.BINARY_OPERATOR,
    ">=": TokenType.BINARY_OPERATOR,
    "&&": TokenType.BINARY_OPERATOR,
    "||": TokenType.BINARY_OPERATOR,
    "<<": TokenType.BINARY_OPERATOR,
    ">>": TokenType.BINARY_OPERATOR,
    "**": TokenType.BINARY_OPERATOR,
    "+=": TokenType.ASSIGNMENT_OPERATOR,
    "-=": TokenType.ASSIGNMENT_OPERATOR,
    "*=": TokenType.ASSIGNMENT_OPERATOR,
    "/=": TokenType.ASSIGNMENT_OPERATOR,
    "%=": TokenType.ASSIGNMENT_OPERATOR,
    "^=": TokenType.ASSIGNMENT_OPERATOR,
    "&=": TokenType.ASSIGNMENT_OPERATOR,
    "|=": TokenType.ASSIGNMENT_OPERATOR,
    "++": TokenType.BINARY_OPERATOR,
    "--": TokenType.BINARY_OPERATOR,
}

KEYWORDS = {
    "let": TokenType.LET,
    "const": TokenType.CONST,
    "func": TokenType.FUNC,
    "if": TokenType.IF,
    "else": TokenType.ELSE,
    "return": TokenType.RETURN,
    "import": TokenType.IMPORT,
    "from": TokenType.FROM,
    "as": TokenType.AS,
    "loop": TokenType.LOOP,
    "while": TokenType.WHILE,
    "for": TokenType.FOR,
    "in": TokenType.IN,
    "break": TokenType.BREAK,
    "continue": TokenType.CONTINUE,
    "del": TokenType.DEL,
    "class": TokenType.CLASS,
    "enum": TokenType.ENUM,
    "assert": TokenType.ASSERT,
}


@dataclass()
class Token:
    """Token class."""

    value: Any
    type: TokenType
    position: tuple[int, int, int]  # (line, col, length)


def is_skipable(
    char: str, current_line: int, current_col: int
) -> Tuple[bool, int, int]:
    """Check if a character is a skipable character."""
    if char in ("\n", "\r"):
        return (True, current_line + 1, 0)  # Skip newlines and carriage returns

    return (
        char
        in (
            " ",
            "\t",
        ),
        current_line,
        current_col,
    )  # Skip spaces and tabs


def pop_src(src: list[str], current_col: int, count: int = 1) -> Tuple[str, int]:
    """Pop the first element of the source code."""
    result = ""
    for _ in range(count):
        result += src.pop(0)
    return result, current_col + len(result)


def tokenize(source_code: str) -> list[Token]:
    """Tokenize the source code."""
    tokens = []  # Initialize the tokens list
    src = list(source_code)
    comment = False  # Comment flag
    current_line = 1
    current_col = 0

    while len(src) > 0:
        negative_num = False  # Reset the negative number flag

        # Skip comments
        if comment:
            if src[0] in ("\n", "\r"):
                current_line += 1  # Increment the line count
                current_col = 0  # Reset the column count
                comment = False
            elif src[0] == ";":
                comment = False
            _, current_col = pop_src(src, current_col)
            continue

        # Skip skipable characters
        skipable, current_line, current_col = is_skipable(
            src[0], current_line, current_col
        )
        if skipable:  # spaces, newlines, tabs, and carriage returns
            _, current_col = pop_src(src, current_col)
            continue

        # Check for double character tokens first
        if len(src) > 1 and src[0] + src[1] in DOUBLE_CHAR_TOKENS:
            token, current_col = pop_src(src, current_col, 2)
            tokens.append(
                Token(token, DOUBLE_CHAR_TOKENS[token], (current_line, current_col, 2))
            )
            continue

        # Check for single character tokens
        if src[0] in SINGLE_CHAR_TOKENS:
            token, current_col = pop_src(src, current_col)

            # Single character token
            tokens.append(
                Token(token, SINGLE_CHAR_TOKENS[token], (current_line, current_col, 1))
            )
            continue

        # Check for comments
        if src[0] == "#":
            comment = True
            _, current_col = pop_src(src, current_col)
            continue

        # If its not a single/double character token, check for negative numbers/variables
        if src[0] == "-":
            if len(src) > 0 and (src[1].isdigit() or src[1].isalpha() or src[1] == "_"):
                negative_num = True  # Set negative number flag
                _, current_col = pop_src(src, current_col)
            else:
                # If its not a negative number, its a "-" operator
                token, current_col = pop_src(src, current_col)
                tokens.append(
                    Token(
                        token,
                        TokenType.BINARY_OPERATOR,
                        (current_line, current_col, 1),
                    )
                )
                continue

        # Check for multi character tokens
        if src[0].isdigit():  # Number
            number, current_col = pop_src(src, current_col)

            if negative_num:
                number = "-" + number  # Add negative sign to the number

            dots = 0
            while len(src) > 0 and (src[0].isdigit() or src[0] == "."):
                if src[0] == ".":
                    dots += 1
                    if dots > 1:
                        break  # Only one dot is allowed in a number
                token, current_col = pop_src(src, current_col)
                number += token
            tokens.append(
                Token(
                    number,
                    TokenType.NUMBER,
                    (
                        current_line,
                        current_col,
                        (len(number) + 1) if negative_num else len(number),
                    ),
                )
            )

        elif src[0].isalpha() or src[0] == "_":  # Identifier
            identifier, current_col = pop_src(src, current_col)
            while len(src) > 0 and (
                src[0].isalpha() or src[0].isdigit() or src[0] == "_"
            ):
                token, current_col = pop_src(src, current_col)
                identifier += token

            if identifier in KEYWORDS:  # Check if the identifier is a keyword
                tokens.append(
                    Token(
                        identifier,
                        KEYWORDS[identifier],
                        (current_line, current_col, len(identifier)),
                    )
                )

            else:  # If its not a keyword, its an identifier
                if negative_num:  # Fake a unary minus operator
                    tokens.append(
                        Token("(", TokenType.OPEN_PAREN, (current_line, current_col, 1))
                    )
                    tokens.append(
                        Token("0", TokenType.NUMBER, (current_line, current_col, 1))
                    )
                    tokens.append(
                        Token(
                            "-",
                            TokenType.BINARY_OPERATOR,
                            (current_line, current_col, 1),
                        )
                    )

                tokens.append(
                    Token(
                        identifier,
                        TokenType.IDENTIFIER,
                        (current_line, current_col, len(identifier)),
                    )
                )

                if negative_num:  # Finish the unary minus operator
                    tokens.append(
                        Token(
                            ")", TokenType.CLOSE_PAREN, (current_line, current_col, 1)
                        )
                    )

        elif src[0] == '"':  # String
            _, current_col = pop_src(src, current_col)  # Remove the opening quote
            string = ""
            while len(src) > 0 and src[0] != '"':
                token, current_col = pop_src(src, current_col)
                string += token
            _, current_col = pop_src(src, current_col)  # Remove the closing quote
            tokens.append(
                Token(
                    string,
                    TokenType.STRING,
                    (current_line, current_col, len(string) + 2),
                )
            )

        else:
            # If this is reached, its an unknown character
            syntax_error(
                source_code,
                (current_line, current_col, 1),
                f"Unknown character found in source '{Fore.MAGENTA}{src.pop(0)}{Fore.RESET}'",
            )

    # Add the final EOF token
    tokens.append(Token("EOF", TokenType.EOF, (current_line, current_col, 1)))

    return tokens



"""Environment class for storing variables (also called scope)."""

import json
import math
import os
import random
import sys
import time
from pathlib import Path
from urllib.parse import quote, unquote

import requests

from eryx.runtime.values import (
    ArrayValue,
    BooleanValue,
    ClassValue,
    EnumValue,
    FunctionValue,
    NativeFunctionValue,
    NullValue,
    NumberValue,
    ObjectValue,
    RuntimeValue,
    StringValue,
)

BUILTINS = {}


# pylint: disable=invalid-name
class Environment:
    """Environment class."""

    def __init__(
        self, parent_env: "Environment | None" = None, disable_file_io: bool = False
    ):
        self.is_global = parent_env is None
        self.parent = parent_env
        self.constants = []
        self.variables = {}
        self.disable_file_io = (
            disable_file_io if not parent_env else parent_env.disable_file_io
        )

        if self.is_global:
            self.setup_scope()

    def declare_variable(
        self,
        variable_name: str,
        value: RuntimeValue,
        constant: bool = False,
        overwrite: bool = False,
    ) -> RuntimeValue:
        """Declare a variable in the current scope."""
        # Raise an exception if the variable is already declared
        if variable_name in self.variables and not overwrite:
            raise RuntimeError(f'Variable "{variable_name}" already declared')

        self.variables[variable_name] = value

        if constant:
            self.constants.append(variable_name)

        return value

    def assign_variable(
        self, variable_name: str, value: RuntimeValue, overwrite: bool = False
    ) -> RuntimeValue:
        """Assign a value to a variable in the current scope."""
        environment = self.resolve(variable_name)

        if variable_name in environment.constants and not overwrite:
            raise RuntimeError(f'Cannot assign to constant variable "{variable_name}"')

        environment.variables[variable_name] = value
        return value

    def lookup_variable(self, variable_name: str) -> RuntimeValue:
        """Lookup a variable in the current scope."""
        environment = self.resolve(variable_name)
        return environment.variables[variable_name]

    def resolve(self, variable_name: str) -> "Environment":
        """Resolve a variable name to an environment."""
        # Return self if variable_name exists in the current scope
        if variable_name in self.variables:
            return self
        # If it does not exist, check the parent scope
        if self.parent:
            return self.parent.resolve(variable_name)
        # If it does not exist in the parent scope, raise an exception
        raise RuntimeError(f'Variable "{variable_name}" not found in scope')

    def delete_variable(self, variable_name: str) -> None:
        """Delete a variable from the current scope."""
        if variable_name in self.variables:
            if variable_name in self.constants:
                del self.constants[self.constants.index(variable_name)]
            del self.variables[variable_name]
        else:
            raise RuntimeError(f'Variable "{variable_name}" not found in scope')

    def setup_scope(self) -> None:
        """Setup the global scope."""
        # Declare global variables
        self.declare_variable("true", BooleanValue(True), True)
        self.declare_variable("false", BooleanValue(False), True)
        self.declare_variable("null", NullValue(), True)

        # Declare native methods
        self.declare_variable("print", NativeFunctionValue(_print), True)
        self.declare_variable("input", NativeFunctionValue(_input), True)
        self.declare_variable("len", NativeFunctionValue(_len), True)
        self.declare_variable("exit", NativeFunctionValue(_exit), True)
        self.declare_variable("str", NativeFunctionValue(_str), True)
        self.declare_variable("int", NativeFunctionValue(_int), True)
        self.declare_variable("bool", NativeFunctionValue(_bool), True)
        self.declare_variable("array", NativeFunctionValue(_array), True)
        self.declare_variable("type", NativeFunctionValue(_type), True)
        self.declare_variable("range", NativeFunctionValue(_range), True)


def get_value(value: RuntimeValue, inside_array: bool = False) -> str:
    """Get the value of a RuntimeValue."""
    result = ""

    if isinstance(value, NullValue):
        result = "null"

    elif isinstance(value, BooleanValue):
        result = str(value.value).lower()

    elif isinstance(value, NumberValue):
        result = (
            str(value.value)
            if int(value.value) != value.value
            else str(int(value.value))
        )

    elif isinstance(value, StringValue):
        if inside_array:
            result = '"' + value.value + '"'
        else:
            result = value.value

    elif isinstance(value, NativeFunctionValue):
        result = f"<native function {value.call.__name__[1:]}>"

    elif isinstance(value, FunctionValue):
        result = f"<function {value.name}>"

    elif isinstance(value, ArrayValue):
        result += "[ "
        for val in value.elements:
            result += f"{get_value(val, inside_array=True)}, "
        result = result[:-2] + " ]"

    elif isinstance(value, ObjectValue):
        result += "{ "
        for key, val in value.properties.items():
            result += f"{key}: {get_value(val, inside_array=True)}, "
        result = result[:-2] + " }"

    elif isinstance(value, ClassValue):
        result = f"{value.name}("
        if value.arguments:
            result += ", ".join(value.arguments)
        result += "){ "
        for key, val in value.methods.items():
            result += f"{key}: {get_value(val, inside_array=True)}, "
        result = result[:-1] + " }"

    elif isinstance(value, EnumValue):
        result = f"{value.name}(" + "{ " + ", ".join(value.values.keys()) + " })"

    else:
        result = str(value)

    return result


# Native functions
def _print(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    values = []
    for arg in args:
        values.append(get_value(arg))
    print(*values)
    return NullValue()


def _range(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 1:
        if isinstance(args[0], NumberValue):
            return ArrayValue([NumberValue(i) for i in range(int(args[0].value))])
    if len(args) == 2:
        if all(isinstance(i, NumberValue) for i in args):
            return ArrayValue(
                [
                    NumberValue(i)
                    for i in range(
                        int(args[0].value),  # type: ignore
                        int(args[1].value),  # type: ignore
                    )
                ]
            )
    if len(args) == 3:
        if all(isinstance(i, NumberValue) for i in args):
            return ArrayValue(
                [
                    NumberValue(i)
                    for i in range(
                        int(args[0].value),  # type: ignore
                        int(args[1].value),  # type: ignore
                        int(args[2].value),  # type: ignore
                    )
                ]
            )
    raise RuntimeError(f"Cannot create range with {args}")


def _input(args: list[RuntimeValue], env: Environment) -> RuntimeValue:
    if env.disable_file_io:
        raise RuntimeError("Input function is disabled")
    if args and isinstance(args[0], StringValue):
        result = input(args[0].value)
    else:
        result = input()
    return StringValue(result)


def _len(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if isinstance(args[0], StringValue):
        return NumberValue(len(args[0].value))

    if isinstance(args[0], ArrayValue):
        return NumberValue(len(args[0].elements))

    if isinstance(args[0], ObjectValue):
        return NumberValue(len(args[0].properties))

    raise RuntimeError(f"Cannot get length of {args[0]}")


def _exit(args: list[RuntimeValue], env: Environment) -> RuntimeValue:
    if env.disable_file_io:
        raise RuntimeError("Exit function is disabled")
    if args and isinstance(args[0], NumberValue):
        sys.exit(int(args[0].value))
    sys.exit(0)


def _str(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return StringValue("")
    return StringValue(get_value(args[0]))


def _int(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], (StringValue, NumberValue)):
        return NumberValue(int(args[0].value))
    raise RuntimeError(f"Cannot convert {args[0]} to int")


def _bool(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return BooleanValue(False)
    if isinstance(args[0], (StringValue, NumberValue, BooleanValue)):
        return BooleanValue(bool(args[0].value))
    if isinstance(args[0], ArrayValue):
        return BooleanValue(bool(args[0].elements))
    if isinstance(args[0], ObjectValue):
        return BooleanValue(bool(args[0].properties))
    raise RuntimeError(f"Cannot convert {args[0]} to bool")


def _array(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 1:
        if isinstance(args[0], StringValue):
            return ArrayValue([StringValue(char) for char in args[0].value])
        if isinstance(args[0], ObjectValue):
            return ArrayValue(list(args[0].properties.values()))
        if isinstance(args[0], ArrayValue):
            return args[0]
    return ArrayValue(args)


def _type(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    return StringValue(type(args[0]).__name__)


# TIME FUNCTIONS


def _time(_: list[RuntimeValue], __: Environment) -> RuntimeValue:
    return NumberValue(time.time())


def _sleep(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing time argument")
    if not isinstance(args[0], NumberValue):
        raise RuntimeError("Time must be a number")
    time.sleep(args[0].value)
    return NullValue()


def _formatTime(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing time argument")
    if not isinstance(args[0], NumberValue):
        raise RuntimeError("Time must be a number")
    return StringValue(
        time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(args[0].value))
    )


def _getTimezoneOffset(_: list[RuntimeValue], __: Environment) -> RuntimeValue:
    return NumberValue(time.timezone)


# FILE FUNCTIONS


def _readFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing filename argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    try:
        with open(args[0].value, "r", encoding="utf8") as file:
            return StringValue(file.read())
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e


def _writeFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing filename or content argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Content must be a string")
    try:
        with open(args[0].value, "w", encoding="utf8") as file:
            file.write(args[1].value)
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e
    return NullValue()


def _appendFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing filename or content argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Content must be a string")
    try:
        with open(args[0].value, "a", encoding="utf8") as file:
            file.write(args[1].value)
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e
    return NullValue()


def _fileExists(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing filename argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    return BooleanValue(Path(args[0].value).exists())


def _deleteFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing filename argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    try:
        Path(args[0].value).unlink()
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e
    return NullValue()


def _copyFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing source or destination argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Source must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Destination must be a string")
    try:
        Path(args[0].value).write_bytes(Path(args[1].value).read_bytes())
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e
    return NullValue()


def _moveFile(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing source or destination argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Source must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Destination must be a string")
    try:
        Path(args[0].value).replace(Path(args[1].value))
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e
    return NullValue()


def _listFiles(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing directory argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Directory must be a string")
    try:
        return ArrayValue(
            [StringValue(str(file)) for file in Path(args[0].value).iterdir()]
        )
    except FileNotFoundError as e:
        raise RuntimeError(f"Directory '{args[0].value}' not found") from e


def _fileSize(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing filename argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Filename must be a string")
    try:
        return NumberValue(Path(args[0].value).stat().st_size)
    except FileNotFoundError as e:
        raise RuntimeError(f"File '{args[0].value}' not found") from e


# HTTP FUNCTIONS


def _getRequest(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing URL argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("URL must be a string")
    try:
        response = requests.get(args[0].value, timeout=10)
        return ObjectValue(
            {
                "data": StringValue(response.content.decode("utf-8")),
                "status": NumberValue(response.status_code),
            }
        )
    except Exception as e:
        raise RuntimeError(f"Error fetching URL '{args[0].value}'") from e


def _postRequest(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing URL or data argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("URL must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Data must be a string")
    try:
        response = requests.post(args[0].value, data=args[1].value, timeout=10)
        return ObjectValue(
            {
                "data": StringValue(response.content.decode("utf-8")),
                "status": NumberValue(response.status_code),
            }
        )
    except Exception as e:
        raise RuntimeError(f"Error fetching URL '{args[0].value}'") from e


def _putRequest(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing URL or data argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("URL must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Data must be a string")
    try:
        response = requests.put(args[0].value, data=args[1].value, timeout=10)
        return ObjectValue(
            {
                "data": StringValue(response.content.decode("utf-8")),
                "status": NumberValue(response.status_code),
            }
        )
    except Exception as e:
        raise RuntimeError(f"Error fetching URL '{args[0].value}'") from e


def _deleteRequest(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing URL argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("URL must be a string")
    try:
        response = requests.delete(args[0].value, timeout=10)
        return ObjectValue(
            {
                "data": StringValue(response.content.decode("utf-8")),
                "status": NumberValue(response.status_code),
            }
        )
    except Exception as e:
        raise RuntimeError(f"Error fetching URL '{args[0].value}'") from e


def _urlEncode(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing data argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Data must be a string")
    return StringValue(quote(args[0].value))


def _urlDecode(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing data argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Data must be a string")
    return StringValue(unquote(args[0].value))


# MATH FUNCTIONS


def _sqrt(args: list[RuntimeValue], _: Environment):
    if not args:
        raise RuntimeError("Missing number value")
    if not isinstance(args[0], NumberValue):
        raise RuntimeError("Input type must be a number")
    return NumberValue(args[0].value ** 0.5)


def _random(_: list[RuntimeValue], __: Environment):
    return NumberValue(random.random())


def _round(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if len(args) == 1:
        if isinstance(args[0], NumberValue):
            return NumberValue(round(args[0].value))
    elif len(args) == 2:
        if isinstance(args[0], NumberValue) and isinstance(args[1], NumberValue):
            return NumberValue(round(args[0].value, int(args[1].value)))
    raise RuntimeError(f"Cannot round {args[0]}")


def _sum(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(sum(i.value for i in args[0].elements))  # type: ignore
    raise RuntimeError(f"Cannot sum {args[0]}")


def _min(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(min(i.value for i in args[0].elements))  # type: ignore
    raise RuntimeError(f"Cannot get min for {args[0]}")


def _max(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], ArrayValue):
        if all(isinstance(i, NumberValue) for i in args[0].elements):
            return NumberValue(max(i.value for i in args[0].elements))  # type: ignore
    raise RuntimeError(f"Cannot get max for {args[0]}")


def _abs(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) == 0:
        return NumberValue(0)
    if isinstance(args[0], NumberValue):
        return NumberValue(abs(args[0].value))
    raise RuntimeError(f"Cannot get abs for {args[0]}")


def _pow(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing base or exponent argument")
    if all(isinstance(i, NumberValue) for i in args):
        return NumberValue(args[0].value ** args[1].value)  # type: ignore
    raise RuntimeError(f"Cannot get pow for {args}")


def _log(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing base or value argument")
    if all(isinstance(i, NumberValue) for i in args):
        return NumberValue(math.log(args[0].value, args[1].value))  # type: ignore
    raise RuntimeError(f"Cannot get log for {args}")


def _log10(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.log10(args[0].value))
    raise RuntimeError(f"Cannot get log10 for {args[0]}")


def _sin(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.sin(args[0].value))
    raise RuntimeError(f"Cannot get sin for {args[0]}")


def _cos(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.sin(args[0].value))
    raise RuntimeError(f"Cannot get cos for {args[0]}")


def _tan(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.sin(args[0].value))
    raise RuntimeError(f"Cannot get tan for {args[0]}")


def _asin(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.asin(args[0].value))
    raise RuntimeError(f"Cannot get asin for {args[0]}")


def _acos(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.asin(args[0].value))
    raise RuntimeError(f"Cannot get acos for {args[0]}")


def _atan(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.asin(args[0].value))
    raise RuntimeError(f"Cannot get atan for {args[0]}")


def _floor(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.floor(args[0].value))
    raise RuntimeError(f"Cannot get floor for {args[0]}")


def _ceil(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.ceil(args[0].value))
    raise RuntimeError(f"Cannot get ceil for {args[0]}")


def _factorial(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing value argument")
    if isinstance(args[0], NumberValue):
        return NumberValue(math.factorial(int(args[0].value)))
    raise RuntimeError(f"Cannot get factorial for {args[0]}")


# STRING FUNCTIONS


def _splitString(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing string or separator argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("String must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Separator must be a string")
    return ArrayValue([StringValue(s) for s in args[0].value.split(args[1].value)])


def _joinStrings(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing array or separator argument")
    if not isinstance(args[0], ArrayValue):
        raise RuntimeError("Array must be an array")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Separator must be a string")
    if not all(isinstance(s, StringValue) for s in args[0].elements):
        raise RuntimeError("Array must contain only strings")
    return StringValue(args[1].value.join([s.value for s in args[0].elements]))  # type: ignore


def _replaceString(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 3:
        raise RuntimeError("Missing string, search or replace argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("String must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Search must be a string")
    if not isinstance(args[2], StringValue):
        raise RuntimeError("Replace must be a string")
    return StringValue(args[0].value.replace(args[1].value, args[2].value))


def _stringContains(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing string or search argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("String must be a string")
    if not isinstance(args[1], StringValue):
        raise RuntimeError("Search must be a string")
    return BooleanValue(args[1].value in args[0].value)


# ARRAY FUNCTIONS


def _push(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing array or value argument")
    if not isinstance(args[0], ArrayValue):
        raise RuntimeError("First argument must be an array")
    args[0].elements.append(args[1])
    return NullValue()


def _pop(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing array argument")
    if not isinstance(args[0], ArrayValue):
        raise RuntimeError("Argument must be an array")
    return args[0].elements.pop()


def _shift(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing array argument")
    if not isinstance(args[0], ArrayValue):
        raise RuntimeError("Argument must be an array")
    return args[0].elements.pop(0)


def _unshift(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if len(args) < 2:
        raise RuntimeError("Missing array or value argument")
    if not isinstance(args[0], ArrayValue):
        raise RuntimeError("First argument must be an array")
    args[0].elements.insert(0, args[1])
    return NullValue()


def _sort(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing array argument")
    if not isinstance(args[0], ArrayValue):
        raise RuntimeError("Argument must be an array")
    if not all(isinstance(i, NumberValue) for i in args[0].elements):
        raise RuntimeError("Array must contain only numbers")
    args[0].elements.sort(key=lambda x: x.value)  # type: ignore
    return NullValue()


# OS FUNCTIONS


def _getCwd(_: list[RuntimeValue], __: Environment) -> RuntimeValue:
    return StringValue(str(Path.cwd()))


def _changeDir(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing directory argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Directory must be a string")
    os.chdir(args[0].value)
    return NullValue()


def _getEnv(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        return ObjectValue(
            {key: StringValue(value) for key, value in os.environ.items()}
        )
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Variable name must be a string")
    return StringValue(os.environ.get(args[0].value, ""))


def _exec(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing command argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Command must be a string")
    command = os.popen(args[0].value)
    return ObjectValue(
        {
            "output": StringValue(command.read()),
            "status": NumberValue(command.close() or 0),
        }
    )


# JSON FUNCTIONS


def json_to_value(data: dict) -> RuntimeValue:
    """Convert a JSON object to a RuntimeValue."""
    if isinstance(data, list):
        return ArrayValue([json_to_value(value) for value in data])
    if isinstance(data, str):
        return StringValue(data)
    if isinstance(data, (int, float)):
        return NumberValue(data)
    if isinstance(data, bool):
        return BooleanValue(data)
    if data is None:
        return NullValue()
    return ObjectValue({key: json_to_value(value) for key, value in data.items()})


def value_to_json(value: RuntimeValue) -> dict | list | str | int | float | bool | None:
    """Convert a RuntimeValue to a JSON object."""
    if isinstance(value, ObjectValue):
        return {key: value_to_json(val) for key, val in value.properties.items()}
    if isinstance(value, ArrayValue):
        return [value_to_json(val) for val in value.elements]
    if isinstance(value, (StringValue, NumberValue, BooleanValue)):
        return value.value
    if isinstance(value, NullValue):
        return None
    raise RuntimeError("Invalid value")


def _jsonParse(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing JSON string argument")
    if not isinstance(args[0], StringValue):
        raise RuntimeError("Argument must be a string")
    try:
        val = json_to_value(json.loads(args[0].value))
        print(val)
        return val
    except Exception as e:
        raise RuntimeError("Invalid JSON string") from e


def _jsonStringify(args: list[RuntimeValue], _: Environment) -> RuntimeValue:
    if not args:
        raise RuntimeError("Missing JSON object argument")
    try:
        return StringValue(json.dumps(value_to_json(args[0])))
    except Exception as e:
        raise RuntimeError("Invalid JSON object") from e


# Declare builtin modules
BUILTINS["file"] = ObjectValue(
    {
        "read": NativeFunctionValue(_readFile),
        "write": NativeFunctionValue(_writeFile),
        "append": NativeFunctionValue(_appendFile),
        "exists": NativeFunctionValue(_fileExists),
        "delete": NativeFunctionValue(_deleteFile),
        "copy": NativeFunctionValue(_copyFile),
        "move": NativeFunctionValue(_moveFile),
        "list": NativeFunctionValue(_listFiles),
        "size": NativeFunctionValue(_fileSize),
    },
    immutable=True,
)

BUILTINS["http"] = ObjectValue(
    {
        "get": NativeFunctionValue(_getRequest),
        "post": NativeFunctionValue(_postRequest),
        "put": NativeFunctionValue(_putRequest),
        "delete": NativeFunctionValue(_deleteRequest),
        "urlencode": NativeFunctionValue(_urlEncode),
        "urldecode": NativeFunctionValue(_urlDecode),
    },
    immutable=True,
)

BUILTINS["math"] = ObjectValue(
    {
        "sum": NativeFunctionValue(_sum),
        "min": NativeFunctionValue(_min),
        "max": NativeFunctionValue(_max),
        "round": NativeFunctionValue(_round),
        "pi": NumberValue(math.pi),
        "sqrt": NativeFunctionValue(_sqrt),
        "random": NativeFunctionValue(_random),
        "abs": NativeFunctionValue(_abs),
        "pow": NativeFunctionValue(_pow),
        "log": NativeFunctionValue(_log),
        "log10": NativeFunctionValue(_log10),
        "sin": NativeFunctionValue(_sin),
        "cos": NativeFunctionValue(_cos),
        "tan": NativeFunctionValue(_tan),
        "asin": NativeFunctionValue(_asin),
        "acos": NativeFunctionValue(_acos),
        "atan": NativeFunctionValue(_atan),
        "floor": NativeFunctionValue(_floor),
        "ceil": NativeFunctionValue(_ceil),
        "e": NumberValue(math.e),
        "factorial": NativeFunctionValue(_factorial),
    },
    immutable=True,
)

BUILTINS["time"] = ObjectValue(
    {
        "time": NativeFunctionValue(_time),
        "sleep": NativeFunctionValue(_sleep),
        "format": NativeFunctionValue(_formatTime),
        "timezone_offset": NativeFunctionValue(_getTimezoneOffset),
    },
    immutable=True,
)

BUILTINS["string"] = ObjectValue(
    {
        "split": NativeFunctionValue(_splitString),
        "join": NativeFunctionValue(_joinStrings),
        "replace": NativeFunctionValue(_replaceString),
        "contains": NativeFunctionValue(_stringContains),
    },
    immutable=True,
)

BUILTINS["array"] = ObjectValue(
    {
        "push": NativeFunctionValue(_push),
        "pop": NativeFunctionValue(_pop),
        "shift": NativeFunctionValue(_shift),
        "unshift": NativeFunctionValue(_unshift),
        "sort": NativeFunctionValue(_sort),
    },
    immutable=True,
)

BUILTINS["os"] = ObjectValue(
    {
        "cwd": NativeFunctionValue(_getCwd),
        "chdir": NativeFunctionValue(_changeDir),
        "env": NativeFunctionValue(_getEnv),
        "exec": NativeFunctionValue(_exec),
    },
    immutable=True,
)

BUILTINS["json"] = ObjectValue(
    {
        "parse": NativeFunctionValue(_jsonParse),
        "stringify": NativeFunctionValue(_jsonStringify),
    },
    immutable=True,
)



"""Eryx to python transpiler."""

from typing import List

from eryx.frontend.ast import (
    ArrayLiteral,
    AssertStatement,
    AssignmentExpression,
    BinaryExpression,
    BreakLiteral,
    CallExpression,
    ClassDeclaration,
    ContinueLiteral,
    DelStatement,
    EnumDeclaration,
    Expression,
    ForStatement,
    FunctionDeclaration,
    Identifier,
    IfStatement,
    ImportStatement,
    LoopStatement,
    MemberExpression,
    NumericLiteral,
    ObjectLiteral,
    ReturnStatement,
    Statement,
    StringLiteral,
    VariableDeclaration,
    WhileStatement,
)
from eryx.frontend.parser import Parser

parser = Parser()

PRINT_OUTPUT = False
IMPORT_ALIASES = {
    "http": "requests",
}
OUT = {}


def cprint(text: str | float | int = "", tabs: int = 0):
    """Print with tabs."""
    if not OUT:
        OUT["out"] = ""
    if PRINT_OUTPUT:
        print(f"{'    ' * tabs}{text}")
    else:
        OUT["out"] += f"{'    ' * tabs}{str(text)}\n"


def convert_value(value: Expression) -> str:
    """Convert value to python."""
    if isinstance(value, ArrayLiteral):
        return f"[{', '.join([convert_value(v) for v in value.elements])}]"

    if isinstance(value, StringLiteral):
        return f'"{value.value}"'

    if isinstance(value, NumericLiteral):
        return (
            str(value.value)
            if int(value.value) != value.value
            else str(int(value.value))
        )

    if isinstance(value, ObjectLiteral):
        result = "{ "
        for prop in value.properties:
            if prop.value:
                result += f'"{prop.key}": {convert_value(prop.value)}, '
            else:
                result += f'"{prop.key}": {prop.key}, '
        return result[:-2] + " }"

    if isinstance(value, Identifier):
        if value.symbol == "true":
            return "True"
        if value.symbol == "false":
            return "False"
        if value.symbol == "null":
            return "None"
        return value.symbol

    if isinstance(value, BinaryExpression):
        return (
            f"{convert_value(value.left)} {value.operator} {convert_value(value.right)}"
        )

    if isinstance(value, CallExpression):
        return (
            f"{convert_value(value.caller)}"
            f"({', '.join([convert_value(arg) for arg in value.arguments])})"
        )

    if isinstance(value, MemberExpression):
        if not value.computed:
            return f"{convert_value(value.object)}.{convert_value(value.property)}"

        return f"{convert_value(value.object)}[{convert_value(value.property)}]"

    return str(value)


def transpile(
    parsed_code: List[Statement], tabs: int = 0, return_value: bool = False
) -> None:
    """Main transpiler function."""

    for node in parsed_code:
        if isinstance(node, FunctionDeclaration):
            cprint(f"def {node.name}({', '.join(node.arguments)}):", tabs)
            if node.body:
                transpile(node.body, tabs + 1)
            else:
                cprint("pass", tabs + 1)
            cprint()

        elif isinstance(node, IfStatement):
            cprint(f"if {convert_value(node.condition)}:", tabs)
            transpile(node.then, tabs + 1)
            if node.else_:
                cprint("else:", tabs)
                transpile([x for x in node.else_ if x], tabs + 1)

        elif isinstance(node, AssertStatement):
            if node.message:
                cprint(f"assert {convert_value(node.condition)}, {node.message}", tabs)
            else:
                cprint(f"assert {convert_value(node.condition)}", tabs)

        elif isinstance(node, AssignmentExpression):
            cprint(
                f"{convert_value(node.assigne)} {node.operator or '='} {convert_value(node.value)}",
                tabs,
            )

        elif isinstance(node, VariableDeclaration):
            if node.value:
                cprint(f"{node.identifier.symbol} = {convert_value(node.value)}", tabs)
            else:
                cprint(f"{node.identifier.symbol}", tabs)

        elif isinstance(node, WhileStatement):
            cprint(f"while {convert_value(node.condition)}:", tabs)
            transpile(node.body, tabs + 1)
            cprint()

        elif isinstance(node, ForStatement):
            cprint(f"for {node.variable} in {node.iterator}:", tabs)
            transpile(node.body, tabs + 1)
            cprint()

        elif isinstance(node, LoopStatement):
            cprint("while True:", tabs)
            transpile(node.body, tabs + 1)
            cprint()

        elif isinstance(node, BreakLiteral):
            cprint("break", tabs)

        elif isinstance(node, ContinueLiteral):
            cprint("continue", tabs)

        elif isinstance(node, ReturnStatement):
            if node.value:
                cprint(f"return {convert_value(node.value)}", tabs)
            else:
                cprint("return", tabs)

        elif isinstance(node, DelStatement):
            cprint(f"del {convert_value(node.identifier)}", tabs)

        elif isinstance(node, CallExpression):
            cprint(
                f"{convert_value(node.caller)}"
                f"({', '.join([convert_value(arg) for arg in node.arguments])})",
                tabs,
            )

        elif isinstance(node, Identifier):
            if node.symbol == "true":
                cprint("True", tabs)
            if node.symbol == "false":
                cprint("False", tabs)
            if node.symbol == "null":
                cprint("None", tabs)
            cprint(node.symbol, tabs)

        elif isinstance(node, ObjectLiteral):
            cprint("{", tabs)
            for prop in node.properties:
                if prop.value:
                    cprint(f'"{prop.key}": {convert_value(prop.value)},', tabs + 1)
                else:
                    cprint(f'"{prop.key}": {prop.key},', tabs + 1)
            cprint("}", tabs)

        elif isinstance(node, NumericLiteral):
            cprint(node.value, tabs)

        elif isinstance(node, StringLiteral):
            cprint(f'"{node.value}"', tabs)

        elif isinstance(node, BinaryExpression):
            cprint(
                f"{convert_value(node.left)} {node.operator} {convert_value(node.right)}",
                tabs,
            )

        elif isinstance(node, MemberExpression):
            if not node.computed:
                cprint(
                    f"{convert_value(node.object)}.{convert_value(node.property)}", tabs
                )
            else:
                cprint(
                    f"{convert_value(node.object)}[{convert_value(node.property)}]",
                    tabs,
                )

        elif isinstance(node, ArrayLiteral):
            cprint(f"[{', '.join([convert_value(v) for v in node.elements])}]", tabs)

        elif isinstance(node, ImportStatement):
            name = node.module
            if name in IMPORT_ALIASES:
                if node.alias:
                    cprint(f"import {IMPORT_ALIASES[name]} as {node.alias}", tabs)
                elif node.names:
                    cprint(
                        f"from {IMPORT_ALIASES[name]} import {', '.join(node.names)}",
                        tabs,
                    )
                else:
                    cprint(f"import {IMPORT_ALIASES[name]} as {name}", tabs)
            elif node.alias:
                cprint(f"import {name} as {node.alias}", tabs)
            elif node.names:
                cprint(f"from {name} import {', '.join(node.names)}", tabs)
            else:
                cprint(f"import {name}", tabs)

        elif isinstance(node, ClassDeclaration):
            cprint(f"class {node.name}:", tabs)
            if node.arguments:
                for method in node.methods:
                    if isinstance(method, AssignmentExpression):
                        if method.value:
                            cprint(
                                f"{convert_value(method.assigne)} "
                                f"= {convert_value(method.value)}",
                                tabs + 1,
                            )
                cprint(
                    "def __init__(self, " + ", ".join(node.arguments) + "):", tabs + 1
                )
                for arg in node.arguments:
                    cprint(f"self.{arg} = {arg}", tabs + 2)
                for method in node.methods:
                    if isinstance(method, AssignmentExpression):
                        if method.value:
                            cprint(
                                f"self.{convert_value(method.assigne)} "
                                f"= {convert_value(method.value)}",
                                tabs + 2,
                            )
            for method in node.methods:
                if isinstance(method, FunctionDeclaration):
                    cprint(
                        f"def {method.name}({', '.join(['self'] + method.arguments)}):",
                        tabs + 1,
                    )
                    if method.body:
                        transpile(method.body, tabs + 2)
                    else:
                        cprint("pass", tabs + 2)
                elif isinstance(method, VariableDeclaration):
                    cprint(f"{method.identifier} = {method.value}", tabs)
            cprint()

        elif isinstance(node, EnumDeclaration):
            cprint(f"class {node.name}:")
            for value in node.values:
                if isinstance(value, Identifier):
                    cprint(f'{value.symbol} = "{value.symbol}"', tabs + 1)

        else:
            print(f"Unkown type: {type(node)}")

    if return_value:
        output = OUT.get("out", "")
        OUT["out"] = ""
        return output[:-1]


if __name__ == "__main__":
    SAMPLE_CODE = """enum Colors {
    green
    black
    cyan
    magenta
}

print(Colors)

print(Colors.green)"""

    parsed = parser.produce_ast(SAMPLE_CODE)
    print(transpile(parsed.body, return_value=True))

    # with open("sample.py", "w", encoding="utf8") as f:
    # f.write(transpile(parsed))
